/*
classifier_enforcement.c : Creates interface that can be called by main event listener

		           It creates cpuset, and add given process in that cpuset

			   When creating new cpuset, it gets the values for different parameters from rule base
			   and set them accordingly.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <syslog.h>		/* syslog() */
#include <errno.h>		/* sys_errlist[] */
#include "userspace_classifier.h"



/*
Write specified data into given file.
*/
int add_value_to_cpuset_file ( char * path_name, char * data )
{
	int fd, len, ret ;

	fd = open (path_name, O_WRONLY|O_CREAT|O_TRUNC, 0666 );
	if (fd == -1 )
	{
		syslog (LOG_INFO, "ERROR: fopen : %s", strerror(errno));
		exit(1)  ;
	}
	
	len = strlen (data);
	ret = write (fd,data, len);
	if (ret == -1 )
	{
		syslog ( LOG_INFO,  "Couldn't write data [%s] in file %s : Error : %s \n", 
				data, path_name, strerror(errno));
		return -1 ;
	} // end if :

	if (len != ret )
	{
		syslog ( LOG_INFO,  "problem=[Couldn't write all data [%s] in file [%s]]", data, path_name);
		return -1 ;
	} // end if :

	close ( fd );
	return 1 ;
}


/*
Add specified process id to given cpuset
*/
int add_task ( char * path_name, int pid )
{
	char buffer[1024] ;
	int len, ret ;

	char filename[500] ;

	strcpy (filename, path_name) ;
	strcat (filename, "/tasks") ;
	sprintf ( buffer, "%d", pid);
	ret = add_value_to_cpuset_file (filename, buffer);
	if ( ret == -1 )
	{
		return -1 ; // return failure
	}
	return 1 ; // return success
}


/*
set newly created cpuset class with values from configuration file
I need to work on this function, as this function currently hardcoding the values
I need to get these values from rule base and set them accordingly.
*/
int set_class_values (struct rule_parse * cpuset_info, char * path_name)
{
	char buffer[1024] ;
	int len, ret ;
	char *class_name, *data_ptr ;
	char filename[500] ;

	// hardcoding the minimal required values for time being.

	class_name = cpuset_info->cpuset_name ;

	// give some values to /cpus list
	strcpy (filename, path_name) ;
	strcat (filename, "/cpus") ;

	
	data_ptr = cpuset_info->cpumask ;
	ret = add_value_to_cpuset_file (filename, data_ptr);
	if ( ret == -1 )
	{
		syslog ( LOG_INFO, "Prob in adding cpus list\n");
		exit (1) ;
	}


	// give some values to /mems list
	strcpy (filename, path_name) ;
	strcat (filename, "/mems") ;

	data_ptr = cpuset_info->memmask ;
	ret = add_value_to_cpuset_file (filename, data_ptr);
	if ( ret == -1 )
	{
		syslog ( LOG_INFO, "Prob in adding mems list\n");
		exit (1) ;
	}
	return 1 ;
	
} // end function : set_class_values

/*
 * int create_dir will create directory of given path.
 * on success it will return 1
 * on failuer it will return 0
 *
 * Note : please pass absolute path to dir...
 * */
int create_dir (struct rule_parse *cpuset_info , char * path )
{
	char *class_name ;
	int ret, error_value ;
	struct stat temp_stat ;


	class_name = cpuset_info->cpuset_name ;
	ret = lstat( path, &temp_stat );
	error_value = errno ;
	if (ret == -1 )
	{
		if ( error_value == ENOENT )
		{
//			syslog ( LOG_INFO,  "It means directory does not exists, so we will create it\n" );
			ret = mkdir (path, 0777 );
			if (ret == -1 )
			{
				syslog (LOG_INFO, "Error: mkdir : %s",strerror(errno));
				return 0 ; // return failure
			}
			else
			{
				syslog ( LOG_INFO,  ", action=[new class %s created]", class_name );
				// new class has been created, so set the rules for this new class
				set_class_values (cpuset_info, path);
				return 1 ; // return success
			}
		}
		else
		{
			// some other problems are there then...
			syslog (LOG_INFO, "Error: lstat : %s",strerror(errno));
			return 0 ; // return failuer
		}
	}
	else
	{
//		syslog ( LOG_INFO,  "Existing class found %s\n", class_name );
		return 1 ; // return success
	}
} // end function : create_dir

#define BUFF_SIZE 500
#define MTAB_FILE "/etc/mtab"

//get_mount_location: To find the location of the filesystem has been passed
char * get_mount_location ( char * filesystem )
{
	FILE *fp ;
	int fd, ret ;
	char buffer[BUFF_SIZE+1] ;
	char *parse_ptr, *filesystem_ptr, *dev_ptr, *mountpoint_ptr, *save_ptr ;
	char delimiter[] = " " ;
	char *final_mount_point ;

	fp = fopen (MTAB_FILE, "r") ;
	if ( fp == NULL )
	{
		syslog ( LOG_INFO,  "Error: Failed in opening file : %s to get information about mount point\n", MTAB_FILE );
		syslog (LOG_INFO, "ERROR: open : %s", strerror(errno));
		// exiting here, because its fatal error
		exit (1); 
	}



	bzero (buffer, sizeof(buffer));
	while ( (int)fgets (buffer, BUFF_SIZE, fp ) != EOF )
	{
		parse_ptr = buffer ;
		// check if string is empty or not
		for (; *parse_ptr ==' ' || *parse_ptr == '\t' ||  *parse_ptr == '\v'  ; ++parse_ptr );
		if (parse_ptr == NULL )
			break ;



		parse_ptr = buffer ;

		dev_ptr = strtok_r (parse_ptr, delimiter, &save_ptr ) ;
		
		parse_ptr = NULL ;
		mountpoint_ptr = strtok_r (parse_ptr, delimiter, &save_ptr ) ;

		filesystem_ptr = strtok_r (parse_ptr, delimiter, &save_ptr ) ;

		if ( strncmp (filesystem, filesystem_ptr, sizeof (filesystem)) == 0 )
		{
			final_mount_point = (char *)malloc (sizeof (mountpoint_ptr)) ;
			strcpy (final_mount_point, mountpoint_ptr);
			return final_mount_point ;
		}

		if (feof (fp))
		{
			break ;
		}
		bzero (buffer, sizeof(buffer));
	} // end while :

	return NULL ;



} // end function : get_mount_location

/*
make path will create string containing absolute path to cpuset class.
its callers responsibility to free the memory after use.
*/
char * make_path (char *mount_point, char * class_name, char *final_path)
{
	int total_length;

//	total_length = strlen(mount_point)+strlen(class_name)+2;
//	final_path = ( char *) malloc (sizeof(total_length)) ;

	strcpy ( final_path, mount_point);
	strcat (final_path, "/");
	strcat ( final_path, class_name );
	return final_path ;

}

/*
This is the interface exposed, which will add the given PID to specified class.
This function will do all that is required to add pid to class.
If requied it will create new cpuset class.

On success, It will return 0
On failuer, it will return -1.
*/
int add_member_to_class (int pid, struct security_context_list * context_info )
{
	char  path_name[500], *class_name ;

	class_name = context_info->cpuset->data->cpuset_name ;

	// making sure that cpuset filesystem is mounted.
	if (mount_point == NULL )
	{
		mount_point = get_mount_location ("cpuset");
		if (mount_point == NULL )
		{
			syslog ( LOG_INFO,  "problem=[cpuset filesystem not mounted]\n" );
			return -1 ; // failure
		}
	}

	make_path (mount_point, class_name, path_name );
//	syslog ( LOG_INFO,  "Path name for class is %s\n", path_name );
	if ( create_dir ( context_info->cpuset->data , path_name) == 0 )
	{
		syslog ( LOG_INFO,  "pid = %d, problem=[not able to create CPUSET %s]\n", pid, path_name );
		return -1 ; // failure
	}


	if ( add_task ( path_name, pid ) == -1 )
	{
		syslog ( LOG_INFO,  "pid = %d, problem=[adding task to class problem]\n", pid );
		return -1 ; // failure
	}
//	syslog ( LOG_INFO,  " pid = %d, aa action=[added]\n", pid);
	return 1 ; // success
} // end function : add_member_to_class 


