#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#include "userspace_classifier.h"


void print_information(struct security_context_list *);

//security_context_file_parser: To parse the classfication_rulebase file
//which contains sSELiunx security context and corresponding CPUSETs

struct security_context_list *security_context_file_parser(char *file_path) 
{
	 FILE *fd;
	 char buff[buffer_size+1];
	 char *parse_ptr, *save_ptr,*tmp;
	 char delimiter[] = " ";
	struct list *cpuset_parser,*cpuset_parser_tmp;
	struct security_context_list *security_context_parser,*security_context_parser_tmp;
	int len ;

	security_context_head = NULL;
	fd = fopen(file_path,"r");
	if(fd == NULL)
	{
		printf("fopen failed\n");
		exit(1);
	}

	while(fgets(buff,buffer_size,fd) != NULL)
	{
		len = strlen (buff);
		if (len > 0 )
		{
			if ( buff[len-1]== '\n')
			{
				buff[len-1] = '\0';
			}
		}
		else
		{
			continue ;
		}
		parse_ptr=buff;
		for (; *parse_ptr ==' ' || *parse_ptr == '\t' ||  *parse_ptr == '\v'  ; ++parse_ptr );
		if (parse_ptr == NULL )
		{
			if( feof (fd))
			{
			     break ;
			}
			continue ;
		}//end of if
	
		// check if its commented 
		if (*parse_ptr == '#' )
		{
			continue ;
		}
		
		// avoid blank lines 
		if (*parse_ptr == '\0' )
		{
			continue ;
		}

		parse_ptr = buff;
		cpuset_parser = head;

		printf("Parsing line [%s]\n",parse_ptr);

		// getting security context as token
                tmp = strtok_r(parse_ptr,delimiter,&save_ptr);
		
		if(security_context_head == NULL)
		{
			// first element of list
			security_context_head = malloc(sizeof(struct security_context_list));
		
			strcpy(security_context_head->security_context_name,tmp);
			parse_ptr = NULL;

			// getting cpuset as token
			tmp = strtok_r(parse_ptr,delimiter,&save_ptr);
			

			// search for this token in cpuset_list
			security_context_head->cpuset = find_cpuset(tmp);
			if ( security_context_head->cpuset == NULL )
			{
				printf ( "\nSorry, cpuset [%s] specified for security context [%s] does not exists", 
					tmp, security_context_head->security_context_name );
				exit (1);
			}
			

			parse_ptr = NULL;

			security_context_head->security_context_nxt = NULL;

		}
		else
		{
			
			security_context_parser = security_context_head;

			// traversing the list to reach last pointer : can be removed.
			while(security_context_parser != NULL)
			{
				security_context_parser_tmp = security_context_parser;
				security_context_parser = security_context_parser->security_context_nxt;
			}
			
			// copy the security context
			security_context_parser = malloc(sizeof(struct security_context_list));
			strcpy (security_context_parser->security_context_name, tmp);
				

			// get the cpuset name
			parse_ptr = NULL;
			tmp =strtok_r(parse_ptr,delimiter,&save_ptr);

//			printf("context = %s and cpuset = %s\n",security_context_parser->security_context_name, tmp);
			// find for this cpuset in cpuset_list
			security_context_parser->cpuset = find_cpuset(tmp);
			if ( security_context_head->cpuset == NULL )
			{
				printf ( "\nSorry, cpuset [%s] specified for security context [%s] does not exists", 
					tmp, security_context_head->security_context_name );
				exit (1);
			}
			 


			parse_ptr=NULL;
			
			// adding this element to linked list.
			security_context_parser->security_context_nxt = NULL;
			security_context_parser_tmp->security_context_nxt = security_context_parser;


		}//end of if 
	}//end of while
	print_information(security_context_head);
	return security_context_head;
}
//find_cpuset: To search cpuset in the list of all cpuset.

struct list *find_cpuset(char *tmp)
{
			struct list *cpuset_parser;
	
			cpuset_parser = head;
			
			if (cpuset_parser ==NULL )
			{
				return NULL ;
			}
			while(cpuset_parser != NULL)
			{
				if(strcmp(cpuset_parser->data->cpuset_name,tmp) == 0)
				{
					break;
				}
				cpuset_parser = cpuset_parser->nxt;
				
			}

			if (cpuset_parser == NULL )
			{
				return NULL ;
			}
			return cpuset_parser;	
}

//print_information function: To print all Selinux security context 
//and CPUSETs with their related fields

void print_information( struct security_context_list *head11)
{
	struct security_context_list *security_context_tmp;
	security_context_tmp = security_context_head;
	struct rule_parse *cpuset_rules;
	printf ("\n===============================================================\n");
	while(security_context_tmp != NULL)
	{
		printf("security_context_name = [%s] ",security_context_tmp->security_context_name);
	         
		printf("cpuset = [%s] ",security_context_tmp->cpuset->data->cpuset_name);
		cpuset_rules = security_context_tmp->cpuset->data;

		printf("cpuset_other_info = [%s] [%s] [%s] [%s] [%s] [%s] [%s]\n",cpuset_rules->cpumask,cpuset_rules->memmask,cpuset_rules->cpu_exclusive,cpuset_rules->mem_exclusive,cpuset_rules->notify_on_release,cpuset_rules->mem_migrate,cpuset_rules->mem_pressure_enabled);

		security_context_tmp = security_context_tmp->security_context_nxt;
		printf ("\n===============================================================\n");
	} // end while :
} // end function :






////////////////////////////////////////////////////

//search_security_context function searches given s_context
//in provided policies with considering '*' as wild-card.
//This wildcard allows you to skip some part of policy matching.
//You are allowed to specify policies in following manner.

//user_t:role_t:type_t cpuset_name1
//user_t:*:type_t	     cpuset_name2
// *:role_t:*	     cpuset_name3
// *:*:*		     cpuset_name4 # This rule matches all.

// Order of rules is important, put specific rules first, and then put generic rules


struct security_context_list * search_security_context ( char * s_context )
{
        int rule_number, is_failed ;
        char *rule_part, *context_part ;
	struct security_context_list *temp1 ;

        if ( s_context == NULL || !(*s_context) ) return NULL ;


        for ( temp1 = security_context_head ; temp1 != NULL ; temp1 = temp1->security_context_nxt )
        {
                rule_part = temp1->security_context_name ;
                context_part = s_context ;
                is_failed = 0 ;

                // if this rule is empty, following rules also should be empty then.
                if (!(*rule_part)) break ; // it means, all rule after this one are empty.
//                printf (  "Rule no %d present\n",rule_number );
//                printf (  "Comparing %s = %s \n", rule_part, s_context);


                // seperate and compare the user.
                if ( *rule_part == '*' )
                {
                        // if it is '*', then its wild character, and dont need to be matched
                        // forword the rule part till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*rule_part) )
                        {
                        	printf  ( "Error in rule of rule base %s \n",
                                       temp1->security_context_name ) ;
                                is_failed = 1 ;
                                continue ;
                        }

                        // forword the context part till next ':'
                        for ( ; *context_part && *context_part != ':' ; context_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*context_part) )
                        {
                              printf ( "Error in matching rule of rule base %s with context %s\n",
                                        temp1->security_context_name, s_context) ;
                                is_failed = 1 ;
                                continue ;
                        }
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++, context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'
                        if (is_failed) continue ;

                        // If any of string terminates unexpectedly, then return with error
                        if ( !(*rule_part && *context_part))
                        {
                                is_failed = 1 ;
                                continue ;
                        }
                } // end if: if '*'

                // move and pass the seperator ':'
                rule_part++ ;
                context_part++ ;


                // seperate and compare the role.
                if ( *rule_part == '*' )
                {
                        // if it is '*', then its wild character, and dont need to be matched
                        // forword the rule part till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*rule_part) )
                        {
                              printf ( "Error in rule of rule base %s \n",
                                       temp1->security_context_name ) ;
                                is_failed = 1 ;
                                continue ;
                        }

                        // forword the context part till next ':'
                        for ( ; *context_part && *context_part != ':' ; context_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*context_part) )
                        {
                              printf ( "Error in matching rule of rule base %s with context %s\n",
                                       temp1->security_context_name, s_context ) ;
                                is_failed = 1 ;
                                continue ;
                        }
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++,context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'
                        // If any of string terminates unexpectedly, then return with error

                        if ( is_failed ) continue ;

                        if ( !(*rule_part && *context_part))
                        {
                                is_failed = 1 ;
                                continue ;
                        }
                }
                // move and pass the seperator ':'
                rule_part++ ;
                context_part++ ;

                // seperate and compare the domain.
                if ( *rule_part == '*' )
                {
                        // wildcard match successfull :-)
                        return  temp1 ;
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++,context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'

                        if (is_failed ) continue ;

                        // match successfull :-)
                        return  temp1 ;
                }
        } // end for : for each rule in rule base.
        return NULL;
}



